#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include "compilerApi.h"
#include "pass.h"

extern int num_procs, num_dims;

void display_align_root(SgStatement *hedr_stmt)
{
//printf("Align Root: \n");
	for (int i=0 ; i< hedr_stmt->numberOfAttributes(Align_Root) ; i++)
	{
		SgSymbol *sym=(SgSymbol*)hedr_stmt->attributeValue(i,Align_Root);
//printf("root: %s\n",sym->identifier());
	}
//printf("Align Root end \n");
}

void display_align_tree(SgStatement *hedr_stmt)
{
	SgSymbol * symbol = hedr_stmt->symbol()->next();
        SgSymbol * sym_son , *sym_par;

 	printf("Align TREE: \n");
        while ((symbol!=NULL)) {
                if (symbol->scope()==hedr_stmt)
                        break;
                symbol=symbol->next();
        }
        if (symbol==NULL) return ;

        for( SgSymbol *symb=symbol ; symb; symb=symb->nextInScope()) {
		if (symb->scope() == NULL ) continue;
                if (symb->numberOfAttributes(Align_Son) >0 )
                {
			printf("++%s %d-%d: ",symb->identifier(),symb->numberOfAttributes(Align_Son),symb->numberOfAttributes(Align_Parent));
			for (int i =0; i<symb->numberOfAttributes(Align_Son) ; i++)
			{
                        	sym_son=(SgSymbol *)symb->attributeValue(i,Align_Son);
				printf("-- %s ",sym_son->identifier());
			}
			printf("\n");
	        }
        }
        printf("Align TERR end \n");
}

// collect variable which is root of the tree
void collect_align_root(SgStatement *hedr_stmt)
{
	SgSymbol * symb = hedr_stmt->symbol()->next();
	SgType    *atype;

	// clear all attributes about Align_Root
	hedr_stmt->deleteAllAttribute(Align_Root);

	while ((symb!=NULL)) {
    		if (symb->scope()==hedr_stmt)
      			break;
    		symb=symb->next();
  	}
	if (symb==NULL) return ;

	for( ; symb; symb=symb->nextInScope()) {
		if (symb->scope() == NULL ) continue;
		if (symb->scope() != hedr_stmt ) continue;
		if (symb->type() == NULL) continue;
		atype =symb->type();
		// scalar not considered
		if (atype->variant() != T_ARRAY ) continue;
		atype=atype->baseType();
		if (atype->variant() == T_PROCESSORS) continue;
		if (symb->numberOfAttributes(Align_Parent) == 0)
			hedr_stmt->addAttribute(Align_Root,symb,sizeof(symb) );
    	}
}

//   fill the symbol with Align_Parent Align_Son
void  collect_align_attributes(SgStatement *s)
{
	SgExpression *align_op,*align_with_op;
	SgSymbol     *align_par,*align_son;
	SgStatement *hedr_stmt = s->getScopeForDeclare();

	align_op = s->expr(2)->lhs();
        align_with_op=align_op->rhs();

        align_par = align_op->rhs()->rhs()->symbol();
        align_son= s->expr(0)->lhs()->symbol();
	
	if (align_par == align_son )
	{
		printf("%s is align to itself !\n",align_par->identifier() );
		s->deleteStmt();
		return;
	}

	align_par->addAttribute(Align_Son , align_son , sizeof(align_son));
	align_son->addAttribute(Align_Parent , align_par , sizeof(align_par));

	align_son->addAttribute(Align_Stmt , s , sizeof(s));

	// if *( ) presents 
	if (align_with_op->lhs())      
		align_son->addAttribute(Align_Star, NULL,0);
}

//   fill the symbol with Dist_Stmt         must be done only once 
void  collect_dist_attributes(SgStatement *s)
{
        SgSymbol *distributee = s->expr(0)->lhs()->symbol();

        distributee->addAttribute(Dist_Stmt , s , sizeof(s));
}

//   fill the symbol with Temp_Stmt         must be done only once 
void  collect_temp_attributes(SgStatement *s)
{
	SgStatement *hedr_stmt = s->getScopeForDeclare();
	for ( SgExpression *expr_list = s->expr(0) ; expr_list ; expr_list=expr_list->rhs())
	{
		SgSymbol *sym = expr_list->lhs()->symbol();
		sym->addAttribute(Temp_Stmt , s , sizeof(s));
		hedr_stmt->addAttribute(Temp_Def , sym , sizeof(sym) );
	}
}

void  collect_alloc_attributes(SgStatement *s)
{
	SgExpression *varlist=s->expr(0);

	for ( ; varlist ; varlist=varlist->rhs() )
	{
	     SgExpression *expr = varlist->lhs();
	     SgSymbol *sym = expr->symbol();
	       // ?????????????
	     if (isSgRecordRefExp(expr))
		sym = isSgRecordRefExp(expr)->lastFieldName();
             ASSERT(sym);
	     sym->addAttribute(Alloc_Stmt,s,sizeof(s));
	       // collect array size and array symbol information
	     sym->addAttribute(Alloc_Size,varlist->lhs(),sizeof(SgExpression*));
	       // standardize 
	     SgArrayRefExp *array = isSgArrayRefExp(expr);
	     if (array) {
		  // standardize array refer as a(:4) => a(1:4)
		for (int i=0; i<array->numberOfSubscripts(); i++)
		{
		  SgExpression *exp = array->subscript(i);
		  if (exp->variant() != DDOT) {
			SgSubscriptExp *newsub = new SgSubscriptExp(*new SgValueExp(1), *exp->copyPtr());
			exp->replace(*newsub);
		  }
		}
	     }
	}
}

//   fill the symbol with Dynamic         must be done only once 
void  collect_dynamic_attributes(SgStatement *s)
{
        SgSymbol *dynSymb = s->expr(0)->lhs()->symbol();

        dynSymb->addAttribute(Dynamic_Stmt , s , sizeof(s));
}

/***************************************************************************
		Get Information of Processors
***************************************************************************/
int processor_size(SgExpression * expr, int &dim)
{
  switch(expr->variant()) {
  case EXPR_LIST: {
    int size1 = processor_size(expr->lhs(), dim);
    int size2 = expr->rhs()?processor_size(expr->rhs(), dim):1;
    return size1 * size2;
    }
    break;

  case DDOT:
    dim++;
    return processor_size(expr->lhs(), dim);
    break;

  case RANGE_OP:
    {
      SgExpression *start = expr->lhs();
      SgExpression *end = expr->rhs();
 
      int s=1,e;
      if (start)
        if (start->isInteger()) s = start->valueInteger();
        else return -1;
      if (end->isInteger()) e = end->valueInteger();
      else return -1;
      return (e-s+1);
    };
    break;

  default:
    printf("Wrong array index in declaration: %d\n", expr->id());
  };
  return 0;
}

void deal_processor_array(SgStatement * s, SgExpression * expr)
{
  switch(expr->variant()) {
  case EXPR_LIST:
    deal_processor_array(s, expr->operand(1));
    if (expr->operand(2))
      deal_processor_array(s, expr->operand(2));
    break;
    
  case ARRAY_REF:
    {
	int dim=0;
	int size = processor_size(expr->lhs(), dim);
	if (size > 0) {
	  num_procs = size;
	  num_dims = dim;
printf("Set default processors as: %d dim %d number.\n",num_dims,num_procs);
	}
    };
    break;

  default:
    printf("Wrong array index in declaration: %d\n", expr->id());
    break;

  };
}

void  collect_proc_attributes(SgStatement *s)
{
	// get the proc_hedr
	SgStatement *hedr_stmt = s->getScopeForDeclare();
	SgExpression *proc = s->expr(0)->lhs();
        SgSymbol     *sym  = proc->symbol();
	hedr_stmt->addAttribute(Proc_Def , sym , sizeof(sym) );
	  // set to global_bfnd
	SgStatement *global = hedr_stmt->controlParent();
        while (global->controlParent()) global = global->controlParent();
        ASSERT(global);
	global->addAttribute(Proc_Def , sym , sizeof(sym) );
	  // try to set default processor
	deal_processor_array(s,  s->expr(0));
}

